#include "mainwindow.h"
#include "ui_mainwindow.h"

#include <QFileDialog>
#include <QMessageBox>
#include <QRegExpValidator>

#include "PlotterWidget.hpp"

#include "utils.hpp"
#include "Objects/Well.hpp"

MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent)
    , ui(new Ui::MainWindow)
{
    ui->setupUi(this);
    initPlot();

    QRegExp rx("-?\\d+\\.?\\d*");
    QValidator * validator = new QRegExpValidator(rx);
    ui->azimuthVs->setValidator(validator);

    mPlotter->setAzimuthVS(getAzimuthRad());
}

void MainWindow::on_importTrajectory_clicked()
{
    QString fileName =  QFileDialog::getOpenFileName(this, "Trajectory Import", "", tr("Json файлы (*.json)"));
    QFile file(fileName);
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
        return;

    QString jsonFile(file.readAll());
    if (jsonFile.isNull())
        return;

    mWell = Utils::parseJson(jsonFile);
    if (mWell == nullptr)
        return;

    Utils::calculateTrajectory(*mWell);
    Utils::calculateVerticalSectionProjection(*mWell, getAzimuthRad());

    mPlotter->drawTrajectory(*mWell);
    mPlotter->scaleByTrajectory(*mWell);
}

void MainWindow::on_importLog_clicked()
{
    QString fileName =  QFileDialog::getOpenFileName(this, "Import Log", "", tr("Json файлы (*.json)"));
    QFile file(fileName);
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
        return;

    QString jsonFile(file.readAll());
    if (jsonFile.isNull())
        return;

    mCurve = Utils::parseLogJson(jsonFile);
    if (mCurve == nullptr)
        return;

    Utils::calculateCurveTrajectory(*mCurve, *mWell);
    Utils::calculateVerticalSectionCurveProjection(*mCurve, getAzimuthRad());

    mPlotter->drawCurveTrajectory(*mCurve);
}

MainWindow::~MainWindow()
{
    delete ui;
}

void MainWindow::on_exit_clicked()
{
    this->close();
}

void MainWindow::on_azimuthVs_returnPressed()
{
    Utils::calculateVerticalSectionProjection(*mWell, getAzimuthRad());
    Utils::calculateVerticalSectionCurveProjection(*mCurve, getAzimuthRad());

    mPlotter->setAzimuthVS(getAzimuthRad());
    mPlotter->drawTrajectory(*mWell);
    mPlotter->drawCurveTrajectory(*mCurve);
    mPlotter->scaleByTrajectory(*mWell);
}

void MainWindow::initPlot()
{
    mPlotter = new PlotterWidget(ui->centralWidget);
    QSizePolicy sizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::MinimumExpanding);
    sizePolicy.setHorizontalStretch(0);
    sizePolicy.setVerticalStretch(0);
    sizePolicy.setHeightForWidth(mPlotter->sizePolicy().hasHeightForWidth());
    mPlotter->setSizePolicy(sizePolicy);

    ui->horizontalLayout->insertWidget(0, mPlotter);
}

double MainWindow::getAzimuthRad() const
{
    return Utils::degreeToRadians(ui->azimuthVs->text().toDouble());
}
